/*
 * DocumentHub v8 - JavaScript module
 */

const axios = require('axios');
const mime = require('mime');

(function (global) {
	var documenthub = (function () {
		return function () {}
	})();


	documenthub.server = 'https://developer.ibm.com/edge/documenthub/v8'
	documenthub.v7Server = 'https://caas7.mybluemix.net/edge/documenthub/api'
	documenthub.ssoServer = 'https://developer.ibm.com/edge/documenthub/sso'
	documenthub.antivirusServer = 'https://documenthub.dev/edge/documenthub/antivirus/api'
	documenthub.authHeader = {}
	documenthub.logs = true
	documenthub.logger = false
	documenthub.cachePeriod = {
		catalogs: 0,
		catalogFiles: 0,
		documents: 0,
		documentFiles: 0,
		search: 0
	}
	documenthub.cache = {}


	// Logs

	documenthub.debug = (logger) => {
		logEntry('debug')
		documenthub.logs = 'debug'
		documenthub.logger = logger
	}

	documenthub.noLogs = () => {
		logEntry('noLogs')
		documenthub.logs = false
		documenthub.logger = false
	}


	// Cache

	documenthub.setCachePeriod = (seconds) => {
		logEntry('setCachePeriod', seconds)
		documenthub.cachePeriod = {
			catalogs: seconds,
			catalogFiles: seconds,
			documents: seconds,
			documentFiles: seconds,
			search: seconds
		}
		documenthub.cache = {}
	}

	documenthub.cacheCatalogs = (seconds) => {
		logEntry('cacheCatalogs', seconds)
		documenthub.cachePeriod.catalogs = seconds
	}

	documenthub.cacheCatalogFiles = (seconds) => {
		logEntry('cacheCatalogFiles', seconds)
		documenthub.cachePeriod.catalogFiles = seconds
	}

	documenthub.cacheDocuments = (seconds) => {
		logEntry('cacheDocuments', seconds)
		documenthub.cachePeriod.documents = seconds
	}

	documenthub.cacheDocumentFiles = (seconds) => {
		logEntry('cacheDocumentFiles', seconds)
		documenthub.cachePeriod.documentFiles = seconds
	}

	documenthub.cacheSearch = (seconds) => {
		logEntry('cacheSearch', seconds)
		documenthub.cachePeriod.search = seconds
	}
	

	// Auth functions

	documenthub.basicAuth = (libraryId, librarySecret, userId) => {
		logEntry('basicAuth', libraryId, '*****', userId)
		if (userId) libraryId += '/' + userId
		documenthub.authHeader = {
			Authorization: "Basic " + (typeof Buffer === 'undefined' ? btoa(libraryId + ':' + librarySecret) : Buffer.from(libraryId + ':' + librarySecret).toString('base64'))
		}
	}

	documenthub.bearerAuth = (token) => {
		logEntry('bearerAuth', token)
		documenthub.authHeader = {
			Authorization: "Bearer " + token
		}
	}

	documenthub.getW3idSSOLoginUrl = (libraryId, queryString) => {
		logEntry('getW3idSSOLoginUrl', libraryId, queryString)
		return documenthub.ssoServer + '/w3id/login/' + libraryId + (queryString ? '?' + queryString : '')
	}

	documenthub.getIBMidSSOLoginUrl = (libraryId, queryString) => {
		logEntry('getIBMidSSOLoginUrl', libraryId, queryString)
		return documenthub.ssoServer + '/ibmid/login/' + libraryId + (queryString ? '?' + queryString : '')
	}

	documenthub.getW3idUser = async (code) => {
		logEntry('getW3idUser', code)
		return await doRequest({
			url: documenthub.ssoServer + '/w3id/user/' + code,
			method: 'get'
		})
	}

	documenthub.getIBMidUser = async (code) => {
		logEntry('getIBMidUser', code)
		return await doRequest({
			url: documenthub.ssoServer + '/ibmid/user/' + code,
			method: 'get'
		})
	}

	documenthub.getTokenByCode = async (code) => {
		logEntry('getTokenByCode', code)
		try {
			let resp = await axios.get(documenthub.v7Server + '/libraries/tokenbycode/' + code)
			return resp.data.token
		} catch (err) {
			let error = {
				status: err.response.status,
				message: err.response.statusText
			}
			logError(error)
			throw error
		}
	}

	documenthub.ensureW3idLogin = async (libraryId, redirectUrl) => {
		logEntry('ensureW3idLogin', libraryId, redirectUrl)
		//use token if exists
		if (documenthub.useTokenCookie()) return

		//use code from url if exists
		if (await documenthub.doLoginCallback()) return

		//redirect to login url
		window.location.href = documenthub.getW3idSSOLoginUrl(libraryId, 'redirect=' + (redirectUrl ? redirectUrl : window.location.href))
	}

	documenthub.ensureIBMidLogin = async (libraryId) => {
		logEntry('ensureIBMidLogin', libraryId)
		//use token if exists
		if (documenthub.useTokenCookie()) return

		//use code from url if exists
		if (await documenthub.doLoginCallback()) return

		//redirect to login url
		window.location.href = documenthub.getIBMidSSOLoginUrl(libraryId, 'redirect=' + window.location.href)
	}

	documenthub.doLoginCallback = async () => {
		logEntry('doLoginCallback')
		//get code
		let urlParams = new URLSearchParams(window.location.search);
		let code = urlParams.get('caasCode')
		log('code', code)
		if (code) {
			//get token
			let token = await documenthub.getTokenByCode(code)

			//set cookie
			documenthub.setTokenCookie(token)

			//set bearer auth
			documenthub.bearerAuth(token)

			//redirect
			let redirect = urlParams.get('redirect')
			log('redirect', redirect)
			if (redirect) window.location.replace(redirect)
			else window.location.replace(removeParameterFromUrl(window.location.href, 'caasCode'))

			return true
		}
		return false
	}

	documenthub.setTokenCookie = (token) => {
		logEntry('setTokenCookie', token)
		var expires = "";
		var date = new Date();
		date.setTime(date.getTime() + (16 * 60 * 60 * 1000));
		expires = "; expires=" + date.toUTCString();
		document.cookie = "caasToken=" + (token || "") + expires + "; path=/";
	}

	documenthub.getTokenCookie = () => {
		logEntry('getTokenCookie')
		try {
			var name = "caasToken"
			var match = document.cookie.match(new RegExp('(^| )' + name + '=([^;]+)'))
			if (match) {
				let token = match[2]
				log('token', token)
				if (token) {
					//verify token expiration
					let user = parseJwt(token)
					if (Date.now() >= user.exp * 1000) throw 'token expired'
					return token
				}
			}
		} catch (err) {
			logError(err)
			documenthub.deleteTokenCookie()
		}
		return null
	}

	documenthub.getDecodedTokenCookie = () => {
		logEntry('getDecodedTokenCookie')
		let token = documenthub.getTokenCookie()
		if (!token) return null
		return parseJwt(token)
	}

	documenthub.getDecodedToken = (token) => {
		logEntry('getDecodedToken')
		return parseJwt(token)
	}

	documenthub.useTokenCookie = () => {
		logEntry('useTokenCookie')
		let token = documenthub.getTokenCookie()
		log('token', token)
		if (token) {
			documenthub.bearerAuth(token)
			return true
		}
		return false
	}

	documenthub.deleteTokenCookie = () => {
		logEntry('deleteTokenCookie')
		document.cookie = "caasToken=; Path=/; Expires=Thu, 01 Jan 1970 00:00:01 GMT;";
	}

	documenthub.getUserACL = async (catalogId) => {
		logEntry('getUserACL', catalogId)
		return await documenthub.get('catalogs/' + catalogId + '/useracl')
	}


	// Generic API functions

	documenthub.get = async (path, server) => {
		logEntry('get', path)
		let config = {
			url: (server ? server : documenthub.server) + '/' + path,
			method: 'get',
			headers: documenthub.authHeader,
		}
		if (path.indexOf('/files?file=') > 0) {
			let ext = path.split('.').pop()
			if (documenthub.binaryExtensions.indexOf(ext) > 0) {
				config.responseType = 'arraybuffer'
			}
		}
		let resp = await doRequest(config)
		return resp
	}

	documenthub.post = async (path, body, contentType, server) => {
		logEntry('post', path, body, contentType)
		let resp = await doRequest({
			url: (server ? server : documenthub.server) + '/' + path,
			method: 'post',
			headers: getRequestHeaders(path, body, contentType),
			data: body
		})
		return resp
	}

	documenthub.put = async (path, body, contentType) => {
		logEntry('put', path, body, contentType)
		let resp = await doRequest({
			url: documenthub.server + '/' + path,
			method: 'put',
			headers: getRequestHeaders(path, body, contentType),
			data: body
		})
		return resp
	}

	documenthub.patch = async (path, body, contentType) => {
		logEntry('patch', path, body, contentType)
		let resp = await doRequest({
			url: documenthub.server + '/' + path,
			method: 'patch',
			headers: getRequestHeaders(path, body, contentType),
			data: body
		})
		return resp
	}

	documenthub.deleteRequest = async (path) => {
		logEntry('deleteRequest', path)
		let resp = await doRequest({
			url: documenthub.server + '/' + path,
			method: 'delete',
			headers: documenthub.authHeader
		})
		return resp
	}

	// Library functions
	documenthub.getLibraries = async () => {
		logEntry('getLibraries')
		throw {
			status: 501,
			message: 'Not Implemented: getLibraries'
		}
	}

	documenthub.getLibrary = async (libraryId) => {
		logEntry('getLibrary', libraryId)
		throw {
			status: 501,
			message: 'Not Implemented: getLibrary'
		}
	}

	documenthub.getLibraryDocuments = async () => {
		logEntry('getLibraryDocuments')
		throw {
			status: 501,
			message: 'Not Implemented: getLibraryDocuments'
		}
	}

	documenthub.searchLibraryDocuments = async (query) => {
		logEntry('searchLibraryDocuments', query)
		return await documenthub.get('search/library/advanced?query=' + encodeURIComponent(JSON.stringify(query)))
	}


	// Catalog functions

	documenthub.getCatalogs = async () => {
		logEntry('getCatalogs')
		return await documenthub.get('catalogs')
	}

	documenthub.getCatalog = async (catalogId) => {
		logEntry('getCatalog', catalogId)
		return await documenthub.get('catalogs/' + catalogId)
	}

	documenthub.getCatalogByRepository = async (repositoryUrl, branch) => {
		logEntry('getCatalogByRepository', repositoryUrl, branch)
		throw {
			status: 501,
			message: 'Not Implemented: getCatalogByRepository'
		}
	}

	documenthub.updateCatalog = async (catalogId, jsonObject) => {
		logEntry('updateCatalog', catalogId, jsonObject)
		return await documenthub.put('catalogs/' + catalogId, jsonObject)
	}

	documenthub.patchCatalog = async (catalogId, jsonObject) => {
		logEntry('patchCatalog', catalogId, jsonObject)
		return await documenthub.patch('catalogs/' + catalogId, jsonObject)
	}

	documenthub.deleteCatalog = async (catalogId) => {
		logEntry('deleteCatalog', catalogId)
		return await documenthub.deleteRequest('catalogs/' + catalogId)
	}

	documenthub.patchSubcatalog = async (catalogId, subcatalogId, jsonObject) => {
		logEntry('patchSubcatalog', catalogId, subcatalogId, jsonObject)
		throw {
			status: 501,
			message: 'Not Implemented: patchSubcatalog'
		}
	}

	documenthub.createGitHubCatalog = async (repositoryUrl, branch, libraryId) => {
		logEntry('createGitHubCatalog', repositoryUrl, branch, libraryId)
		throw {
			status: 501,
			message: 'Not Implemented: createGitHubCatalog'
		}
	}


	// Catalog files functions

	documenthub.getCatalogFiles = async (catalogId, filePath) => {
		logEntry('getCatalogFiles', catalogId, filePath)
		return await documenthub.get('catalogs/' + catalogId + '/files')
	}

	documenthub.getCatalogFile = async (catalogId, filePath) => {
		logEntry('getCatalogFile', catalogId, filePath)
		return await documenthub.get('catalogs/' + catalogId + '/files?file=' + filePath)
	}

	documenthub.addCatalogFile = async (catalogId, filePath, content, contentType) => {
		logEntry('addCatalogFile', catalogId, filePath, content, contentType)
		return await documenthub.post('catalogs/' + catalogId + '/files?file=' + filePath, bodyToBase64(filePath, content))
	}

	documenthub.updateCatalogFile = async (catalogId, filePath, content, contentType) => {
		logEntry('updateCatalogFile', catalogId, filePath, content, contentType)
		return await documenthub.put('catalogs/' + catalogId + '/files?file=' + filePath, bodyToBase64(filePath, content))
	}

	documenthub.patchCatalogFile = async (catalogId, filePath, jsonObject) => {
		logEntry('patchCatalogFile', catalogId, filePath, jsonObject)
		throw {
			status: 501,
			message: 'Not Implemented: patchCatalogFile'
		}
	}

	documenthub.deleteCatalogFile = async (catalogId, filePath) => {
		logEntry('deleteCatalogFile', catalogId, filePath)
		return await documenthub.deleteRequest('catalogs/' + catalogId + '/files?file=' + filePath)
	}



	// Document functions

	documenthub.getDocuments = async (catalogId) => {
		logEntry('getDocuments', catalogId)
		return await documenthub.get('catalogs/' + catalogId + '/documents')
	}

	documenthub.getDocumentsBrief = async (catalogId, fields) => {
		logEntry('getDocumentsBrief', catalogId, fields)
		throw {
			status: 501,
			message: 'Not Implemented: getDocumentsBrief'
		}
	}

	documenthub.getDocument = async (catalogId, documentId, language) => {
		logEntry('getDocument', catalogId, documentId, language)
		return await documenthub.get('catalogs/' + catalogId + '/documents/' + documentId + l(language))
	}

	documenthub.addDocument = async (catalogId, documentId, language, jsonObject) => {
		logEntry('addDocument', catalogId, documentId, language, jsonObject)
		return await documenthub.post('catalogs/' + catalogId + '/documents/' + documentId + l(language), jsonObject, 'application/json')
	}

	documenthub.updateDocument = async (catalogId, documentId, language, jsonObject) => {
		logEntry('updateDocument', catalogId, documentId, language, jsonObject)
		return await documenthub.put('catalogs/' + catalogId + '/documents/' + documentId + l(language), jsonObject, 'application/json')
	}

	documenthub.patchDocument = async (catalogId, documentId, language, jsonObject) => {
		logEntry('patchDocument', catalogId, documentId, language, jsonObject)
		return await documenthub.patch('catalogs/' + catalogId + '/documents/' + documentId + l(language), jsonObject, 'application/json')
	}

	documenthub.deleteDocument = async (catalogId, documentId, language) => {
		logEntry('deleteDocument', catalogId, documentId, language)
		return await documenthub.deleteRequest('catalogs/' + catalogId + '/documents/' + documentId + l(language))
	}

	documenthub.copyDocument = async (catalogId, documentId, language, toCatalogId, toDocumentId, toLanguage) => {
		logEntry('copyDocument', catalogId, documentId, language, toCatalogId, toDocumentId, toLanguage)
		if (!language) language = 'NA'
		if (!toLanguage) toLanguage = 'NA'
		return await documenthub.post('catalogs/' + catalogId + '/documents/' + documentId + '/languages/' + language + '/copy/' + toCatalogId + '/' + toDocumentId + '/' + toLanguage)
	}

	documenthub.getDocumentBatch = async (catalogId, arrayOfIds) => {
		logEntry('getDocumentBatch', catalogId, arrayOfIds)
		return await documenthub.post('catalogs/' + catalogId + '/documentbatch', arrayOfIds)
	}

	documenthub.addDocumentBatch = async (catalogId, jsonObject) => {
		logEntry('addDocumentBatch', catalogId, jsonObject)
		return await documenthub.post('catalogs/' + catalogId + '/documentbatch', jsonObject)
	}

	documenthub.deleteDocumentBatch = async (catalogId, jsonObject) => {
		logEntry('deleteDocumentBatch', catalogId, jsonObject)
		return await documenthub.deleteRequest('catalogs/' + catalogId + '/documentbatch', jsonObject)
	}

	documenthub.addDocumentSet = async (catalogId, jsonObject) => {
		logEntry('addDocumentSet', catalogId, jsonObject)
		return await documenthub.addDocumentsBatch(catalogId, jsonObject)
	}

	documenthub.deleteDocumentSet = async (catalogId, jsonObject) => {
		logEntry('deleteDocumentSet', catalogId, jsonObject)
		return await documenthub.deleteDocumentsBatch(catalogId, jsonObject)
	}


	// Documents search and filter functions

	documenthub.search = async (catalogId, query) => {
		logEntry('search', catalogId, query)
		if (typeof query === 'string' || query instanceof String) {
			return documenthub.searchDocuments(catalogId, query)
		} else {
			//send get requests
			return await documenthub.get('search/catalogs/' + catalogId + '/advanced?query=' + encodeURIComponent(JSON.stringify(query)))
		}
	}

	documenthub.searchDocuments = async (catalogId, text) => {
		logEntry('searchDocuments', catalogId, text)
		return await documenthub.get('search/catalogs/' + catalogId + '/simple/' + text)
	}

	documenthub.advancedSearch = async (catalogId, query) => {
		logEntry('advancedSearch', catalogId, query)
		return await documenthub.post('search/catalogs/' + catalogId + '/advanced', query, 'application/json')
	}

	documenthub.getSearchSuggestions = async (catalogId, text, limit) => {
		logEntry('getSearchSuggestions', catalogId, text, limit)
		return await documenthub.get('search/catalogs/' + catalogId + '/suggestions/' + text + '/' + limit)
	}

	documenthub.getFiltersCount = async (catalogId) => {
		logEntry('getFiltersCount', catalogId)
		return await documenthub.get('search/catalogs/' + catalogId + '/filters')
	}


	// Contents functions

	documenthub.getDocumentContentsToc = async (catalogId, documentId, language) => {
		logEntry('getDocumentContentsToc', catalogId, documentId, language)
		return await documenthub.get('catalogs/' + catalogId + '/documents/' + documentId + l(language) + '/files?file=toc.json')
	}

	documenthub.getDocumentContents = async (catalogId, documentId, language) => {
		logEntry('getDocumentContents', catalogId, documentId, language)
		return await documenthub.get('catalogs/' + catalogId + '/documents/' + documentId + l(language) + '/contents/all')
	}

	documenthub.getDocumentContentsByIndex = async (catalogId, documentId, language, index) => {
		logEntry('getDocumentContentsByIndex', catalogId, documentId, language, index)
		throw {
			status: 501,
			message: 'Not Implemented: getDocumentContentsByIndex'
		}
	}

	documenthub.getDocumentContentAsPage = async (catalogId, documentId, language) => {
		logEntry('getDocumentContentAsPage', catalogId, documentId, language)
		return await documenthub.get('catalogs/' + catalogId + '/documents/' + documentId + l(language) + '/contents/page')
	}


	// Document files functions

	documenthub.getDocumentFiles = async (catalogId, documentId, language) => {
		logEntry('getDocumentFiles', catalogId, documentId, language)
		return await documenthub.get('catalogs/' + catalogId + '/documents/' + documentId + l(language) + '/files')
	}

	documenthub.getDocumentFile = async (catalogId, documentId, language, filePath) => {
		logEntry('getDocumentFile', catalogId, documentId, language, filePath)
		return await documenthub.get('catalogs/' + catalogId + '/documents/' + documentId + l(language) + '/files?file=' + filePath)
	}

	documenthub.addDocumentFile = async (catalogId, documentId, language, filePath, content, contentType) => {
		logEntry('addDocumentFile', catalogId, documentId, language, filePath, content, contentType)
		return await documenthub.post('catalogs/' + catalogId + '/documents/' + documentId + l(language) + '/files?file=' + filePath, bodyToBase64(filePath, content))
	}

	documenthub.updateDocumentFile = async (catalogId, documentId, language, filePath, content, contentType) => {
		logEntry('updateDocumentFile', catalogId, documentId, language, filePath, content, contentType)
		return await documenthub.put('catalogs/' + catalogId + '/documents/' + documentId + l(language) + '/files?file=' + filePath, bodyToBase64(filePath, content))
	}

	documenthub.patchDocumentFile = async (catalogId, documentId, language, filePath, jsonObject) => {
		logEntry('patchDocumentFile', catalogId, documentId, language, filePath, jsonObject)
		throw {
			status: 501,
			message: 'Not Implemented: patchDocumentFile'
		}
	}

	documenthub.deleteDocumentFile = async (catalogId, documentId, language, filePath) => {
		logEntry('deleteDocumentFile', catalogId, documentId, language, filePath)
		return await documenthub.deleteRequest('catalogs/' + catalogId + '/documents/' + documentId + l(language) + '/files?file=' + filePath)
	}

	documenthub.mergeDocumentFile = async (catalogId, documentId, language, filePath, jsonObject) => {
		logEntry('mergeDocumentFile', catalogId, documentId, language, filePath, jsonObject)
		throw {
			status: 501,
			message: 'Not Implemented: mergeDocumentFile'
		}
	}

	documenthub.getBase64DocumentFiles = async (catalogId, documentId, language) => {
		logEntry('getBase64DocumentFiles', catalogId, documentId, language)
		return await documenthub.get('catalogs/' + catalogId + '/documents/' + documentId + l(language) + '/files/base64')
	}

	documenthub.getBase64DocumentFile = async (catalogId, documentId, language, filePath) => {
		logEntry('getBase64DocumentFile', catalogId, documentId, language, filePath)
		return await documenthub.get('catalogs/' + catalogId + '/documents/' + documentId + l(language) + '/files/base64?file=' + filePath)
	}

	// Other functions

	documenthub.sendEmailFromIBM = async (from, to, subject, message) => {
		logEntry('sendEmailFromIBM', from, to, subject, message)
		let jsonObject = {
			from: from,
			to: to,
			subject: subject,
			message: message
		}
		return await documenthub.post('email/ibm', jsonObject, 'application/json', documenthub.v7Server)
	}

	documenthub.antivirusVersion = async () => {
		logEntry('antivirusVersion')
		return await documenthub.get('version', documenthub.antivirusServer)
	}

	documenthub.antivirusScan = async (content) => {
		logEntry('antivirusScan', content.length)
		return await documenthub.post('scan', content, 'application/octet-stream', documenthub.antivirusServer)
	}

	documenthub.binaryExtensions = [
		"3dm",
		"3ds",
		"3g2",
		"3gp",
		"7z",
		"a",
		"aac",
		"adp",
		"ai",
		"aif",
		"aiff",
		"alz",
		"ape",
		"apk",
		"appimage",
		"ar",
		"arj",
		"asf",
		"au",
		"avi",
		"bak",
		"baml",
		"bh",
		"bin",
		"bk",
		"bmp",
		"btif",
		"bz2",
		"bzip2",
		"cab",
		"caf",
		"cgm",
		"class",
		"cmx",
		"cpio",
		"cr2",
		"cur",
		"dat",
		"dcm",
		"deb",
		"dex",
		"djvu",
		"dll",
		"dmg",
		"dng",
		"doc",
		"docm",
		"docx",
		"dot",
		"dotm",
		"dra",
		"DS_Store",
		"dsk",
		"dts",
		"dtshd",
		"dvb",
		"dwg",
		"dxf",
		"ecelp4800",
		"ecelp7470",
		"ecelp9600",
		"egg",
		"eol",
		"eot",
		"epub",
		"exe",
		"f4v",
		"fbs",
		"fh",
		"fla",
		"flac",
		"flatpak",
		"fli",
		"flv",
		"fpx",
		"fst",
		"fvt",
		"g3",
		"gh",
		"gif",
		"graffle",
		"gz",
		"gzip",
		"h261",
		"h263",
		"h264",
		"icns",
		"ico",
		"ief",
		"img",
		"ipa",
		"iso",
		"jar",
		"jpeg",
		"jpg",
		"jpgv",
		"jpm",
		"jxr",
		"key",
		"ktx",
		"lha",
		"lib",
		"lvp",
		"lz",
		"lzh",
		"lzma",
		"lzo",
		"m3u",
		"m4a",
		"m4v",
		"mar",
		"mdi",
		"mht",
		"mid",
		"midi",
		"mj2",
		"mka",
		"mkv",
		"mmr",
		"mng",
		"mobi",
		"mov",
		"movie",
		"mp3",
		"mp4",
		"mp4a",
		"mpeg",
		"mpg",
		"mpga",
		"mxu",
		"nef",
		"npx",
		"numbers",
		"nupkg",
		"o",
		"odp",
		"ods",
		"odt",
		"oga",
		"ogg",
		"ogv",
		"otf",
		"ott",
		"pages",
		"pbm",
		"pcx",
		"pdb",
		"pdf",
		"pea",
		"pgm",
		"pic",
		"png",
		"pnm",
		"pot",
		"potm",
		"potx",
		"ppa",
		"ppam",
		"ppm",
		"pps",
		"ppsm",
		"ppsx",
		"ppt",
		"pptm",
		"pptx",
		"psd",
		"pya",
		"pyc",
		"pyo",
		"pyv",
		"qt",
		"rar",
		"ras",
		"raw",
		"resources",
		"rgb",
		"rip",
		"rlc",
		"rmf",
		"rmvb",
		"rpm",
		"rtf",
		"rz",
		"s3m",
		"s7z",
		"scpt",
		"sgi",
		"shar",
		"snap",
		"sil",
		"sketch",
		"slk",
		"smv",
		"snk",
		"so",
		"stl",
		"suo",
		"sub",
		"swf",
		"tar",
		"tbz",
		"tbz2",
		"tga",
		"tgz",
		"thmx",
		"tif",
		"tiff",
		"tlz",
		"ttc",
		"ttf",
		"txz",
		"udf",
		"uvh",
		"uvi",
		"uvm",
		"uvp",
		"uvs",
		"uvu",
		"viv",
		"vob",
		"war",
		"wav",
		"wax",
		"wbmp",
		"wdp",
		"weba",
		"webm",
		"webp",
		"whl",
		"wim",
		"wm",
		"wma",
		"wmv",
		"wmx",
		"woff",
		"woff2",
		"wrm",
		"wvx",
		"xbm",
		"xif",
		"xla",
		"xlam",
		"xls",
		"xlsb",
		"xlsm",
		"xlsx",
		"xlt",
		"xltm",
		"xltx",
		"xm",
		"xmind",
		"xpi",
		"xpm",
		"xwd",
		"xz",
		"z",
		"zip",
		"zipx"
	]

	// Internal functions 


	function logEntry(func, param1, param2, param3, param4, param5, param6, param7) {
		if (documenthub.logs) {
			var str = func + "(";
			if (param1 !== undefined) str += stringify(param1);
			if (param2 !== undefined) str += ", " + stringify(param2);
			if (param3 !== undefined) str += ", " + stringify(param3);
			if (param4 !== undefined) str += ", " + stringify(param4);
			if (param5 !== undefined) str += ", " + stringify(param5);
			if (param6 !== undefined) str += ", " + stringify(param6);
			if (param7 !== undefined) str += ", " + stringify(param7);
			str += ")";
			log(str);
		}
	}

	function stringify(param) {
		if (typeof param !== 'object') return param
		else {
			let str = JSON.stringify(param)
			let p0 = str.indexOf('{"Authorization":"Basic ')
			if (p0>=0) {
				let p1 = str.indexOf('"', p0+24)
				str = str.substring(0, p0+24) + '*****' + str.substring(p1)
			}
			return str
		}
	}

	function log(name, value) {
		if (documenthub.logs === 'debug') {
			let str = new Date().toISOString().replace(/T/, " ").substring(0, 23)
			str += " DocumentHub: "
			if (value !== undefined) str += name + '=' + stringify(value)
			else str += name
			if (documenthub.logger) documenthub.logger.debug(str)
			else console.log(str)
		}
	}

	function logError(e) {
		if (documenthub.logs) {
			let str = new Date().toISOString().replace(/T/, " ").substring(0, 23)
			str += " DocumentHub: Error: " + stringify(e)
			if (documenthub.logger) documenthub.logger.error(str)
			else console.error(str)
		}
	}

	function l(language) {
		if (language)
			if (typeof language === 'string' && language.length > 0)
				return '/languages/' + language
		return ''
	}

	async function doRequest(config) {
		log('request', config)
		let retry = 2
		while (retry-- >= 0) {
			try {
				let resp = await axios.request(config)
				log('response', resp.status + ' ' + resp.statusText)
				return resp.data
			} catch (err) {
				if (err.response) {
					if (err.response.status === 502 && retry >= 0) {
						logError('Got 502 response. Retrying...')
						await sleepAsync(5000)
					} else {
						//if response.data is Buffer or ArrayBuffer, convert it
						if (isBuffer(err.response.data) || isArrayBuffer(err.response.data)) {
							var utf8decoder = new TextDecoder()
							err.response.data = utf8decoder.decode(err.response.data)
							try {
								err.response.data = JSON.parse(err.response.data)
							} catch(e) {}
						}
		
						if (err.response.status !== 404) {
							delete err.response.request
							delete err.response.config
							logError(err.response)
						}
						let error = {
							status: err.response.status,
							message: err.response.statusText
						}
						if (err.response.data) {
							if (err.response.data.message) {
								error.message = err.response.data.message
								error.details = err.response.data.details
							} else error.message = err.response.data
						}
						if (error.status === 403 && error.message && error.message.indexOf('<TITLE>Access Denied</TITLE>')>0) {
							logError('Got 403 response with access denied message. Retrying...')
							await sleepAsync(5000)
						} else {
							logError(error)
							throw error	
						}
					}
				} else {
					logError(err)
					throw err
				}
			}
		}
	}

	function getRequestHeaders(path, body, contentType) {
		let headers = documenthub.authHeader
		if (contentType) headers["Content-Type"] = contentType
		else if (typeof body === 'string') headers["Content-Type"] = 'text/plain'
		else {
			let p = path.lastIndexOf('.')
			if (p > 0 && p > path.length - 6)
				headers["Content-Type"] = mime.getType(path.substr(p + 1))
			if (!headers["Content-Type"])
				if (typeof body === 'object') headers["Content-Type"] = 'application/json'
		}
		return headers
	}

	function removeParameterFromUrl(url, parameter) {
		return url
			.replace(new RegExp('[?&]' + parameter + '=[^&#]*(#.*)?$'), '$1')
			.replace(new RegExp('([?&])' + parameter + '=[^&]*&'), '$1');
	}

	function parseJwt(token) {
		var base64Url = token.split('.')[1];
		var base64 = base64Url.replace(/-/g, '+').replace(/_/g, '/');
		var jsonPayload = decodeURIComponent(atob(base64).split('').map(function (c) {
			return '%' + ('00' + c.charCodeAt(0).toString(16)).slice(-2);
		}).join(''));

		return JSON.parse(jsonPayload);
	}

	function bodyToBase64(path, body) {
		if (typeof body === 'object') {
			let p = path.lastIndexOf('.')
			if (p > 0 && p > path.length - 6) {
				let ext = path.substr(p + 1)
				if (ext === 'json') body = JSON.stringify(body, null, 2)
			}
		}
		let b64 = (typeof Buffer === 'undefined' ? btoa(body) : Buffer.from(body).toString('base64'))
		return {
			content: b64
		}
	}

	function sleepAsync(ms) {
		return new Promise(resolve => {
			setTimeout(resolve, ms)
		})
	}

	function isBuffer(value) {
		return typeof Buffer !== 'undefined' && Buffer.isBuffer(value)
	}

	function isArrayBuffer(value) {
		return value && value instanceof ArrayBuffer && value.byteLength !== undefined;
	}

	/* global define */
	if (typeof define === 'function' && define.amd) {
		define(function () {
			return documenthub;
		});
	} else if (typeof exports === 'object') {
		module.exports = documenthub;
	} else {
		global.documenthub = documenthub;
	}
})(this);